
/*
	$Id: magictun.c,v 1.13 2003/01/18 11:04:05 anight Exp $

	Copyright (C) 2003 Andrei Nigmatulin <anight at mail dot ru>
	This  program  is free  software;  you can  redistribute  it and/or
	modify  it under the terms of the GNU  General  Public  License  as
	published by the Free Software  Foundation; either version 2 of the
	License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful, but
	WITHOUT  ANY  WARRANTY;   without  even  the  implied  warranty  of
	MERCHANTABILITY  or FITNESS FOR A  PARTICULAR  PURPOSE. See the GNU
	General Public License for more details.
	
	You should have received a copy of the GNU General  Public  License
	along  with  this  program;  if not,  write  to the  Free  Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>

#define __USE_GNU
#include <dlfcn.h>			/* dlsym */

#include <sys/time.h>		/* select */
#include <sys/types.h>		/* socket */
#include <sys/socket.h>		/* socket */
#include <netinet/in.h>		/* in_addr */
#include <arpa/inet.h>		/* inet_addr */
#include <stdlib.h>			/* exit */
#include <string.h>			/* strchr */
#include <netdb.h>			/* gethostbyname */
#include <unistd.h>			/* dup */

#include <errno.h>			/* return codes */

#define BUFSIZE				128
#define PROXY_TIMEOUT		30

#define IPQUAD(sin_addr) \
	( (sin_addr).s_addr >> 0 ) & 0xff, \
	( (sin_addr).s_addr >> 8 ) & 0xff, \
	( (sin_addr).s_addr >> 16 ) & 0xff, \
	( (sin_addr).s_addr >> 24 ) & 0xff

typedef struct _mt_struct mt_struct;

struct _mt_struct {
	struct sockaddr_in orig;
	struct sockaddr_in proxy;
	mt_struct *next;
};

static mt_struct *mt_list = NULL;

static void init (void)
	__attribute__ ((constructor));

static void init (void)
{
	char *home;
	char *rc_name;
	char *buf;
	FILE *rc;
	mt_struct *last;

	if ((home = getenv("HOME")) == NULL) {
#if DEBUG
		fprintf(stderr, "magictun: do not cheat with me ;-)\n");
#endif
		return;
	}

	if ((rc_name = (char*) malloc(strlen(home) + 50)) == NULL) {
#if DEBUG
		fprintf(stderr, "magictun: malloc() failed\n");
#endif
		return;
	}

	if ((buf = (char*) malloc(BUFSIZE)) == NULL) {
#if DEBUG
		fprintf(stderr, "magictun: malloc() failed\n");
#endif
		free(rc_name);
		return;
	}

	sprintf(rc_name, "%s/.magictun", home);

	if ((rc = fopen(rc_name, "r")) == NULL) {
#if DEBUG
		fprintf(stderr, "magictun: unable to open '%s'\n", rc_name);
#endif
		free(rc_name);
		free(buf);
		return;
	}

	free(rc_name);

	do {
		char *ptr1, *ptr2;
		struct sockaddr_in orig, proxy;
		mt_struct *new;
		struct hostent *hostinfo;

		if (fgets(buf, BUFSIZE, rc) == NULL) break;

		ptr1 = strchr(buf, ':');
		if (!ptr1) continue;
		*ptr1++ = 0;
		orig.sin_addr.s_addr = inet_addr(buf);
		if (orig.sin_addr.s_addr == -1) {
			hostinfo = gethostbyname(buf);
			if (hostinfo == NULL) {
				fprintf(stderr, "magictun: unknown host '%s'\n", buf);
				continue;
			}
			orig.sin_addr = *(struct in_addr *) hostinfo->h_addr;
		}
		for (ptr2 = ptr1; *ptr1 && (*ptr1 != ' ' && *ptr1 != '\t'); ptr1++);
		if (!*ptr1) continue;
		*ptr1++ = 0;
		orig.sin_port = htons(atoi(ptr2) & 0xffff);
		for (ptr2 = ptr1; *ptr1 && (*ptr1 == ' ' || *ptr1 == '\t'); ptr1++);
		if (!*ptr1) continue;
		ptr2 = strchr(ptr1, ':');
		if (!*ptr2) continue;
		*ptr2++ = 0;
		proxy.sin_addr.s_addr = inet_addr(ptr1);
		if (proxy.sin_addr.s_addr == -1) {
			hostinfo = gethostbyname(ptr1);
			if (hostinfo == NULL) {
				fprintf(stderr, "magictun: unknown host '%s'\n", ptr1);
				continue;
			}
			proxy.sin_addr = *(struct in_addr *) hostinfo->h_addr;
		}
		for (ptr1 = ptr2; *ptr2 && (*ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\n'); ptr2++);
		*ptr2 = 0;
		proxy.sin_port = htons(atoi(ptr1) & 0xffff);

		new = (mt_struct *) malloc(sizeof(mt_struct));
		if (new == NULL) {
#if DEBUG
			fprintf(stderr, "magictun: malloc() failed\n");
#endif
			free(buf);
			fclose(rc);
			return;
		}
		new->orig = orig;
		new->proxy = proxy;
		new->next = NULL;

#if DEBUG
		fprintf(stderr, "magictun: added %d.%d.%d.%d:%d -> %d.%d.%d.%d:%d\n",
			IPQUAD(orig.sin_addr), ntohs(orig.sin_port),
			IPQUAD(proxy.sin_addr), ntohs(proxy.sin_port)
		);
#endif

		if (mt_list == NULL) mt_list = new;
		else last->next = new;
		last = new;
	} while (1);

	free(buf);
	fclose(rc);
}

int connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
	static int (*old_connect)() = 0;
	mt_struct *a;
	int result;

	if (old_connect == 0) {
		old_connect = (int (*)()) dlsym(RTLD_NEXT, "connect");
		if (old_connect == NULL) {
			fprintf(stderr, "dlsym: %s\n", dlerror());
			return ENOSYS;
		}
	}

	if (serv_addr->sa_family == AF_INET) {
		const struct sockaddr_in *sin = (struct sockaddr_in *) serv_addr;
#if DEBUG
		fprintf(stderr, "magictun: got connect: %d.%d.%d.%d:%d\n",
				IPQUAD(sin->sin_addr), htons(sin->sin_port)
		);
#endif
		for (a = mt_list; a; a = a->next) {
			if (sin->sin_addr.s_addr == a->orig.sin_addr.s_addr
								&& sin->sin_port == a->orig.sin_port) {
				struct sockaddr proxy = *serv_addr;
				struct timeval tv;
				fd_set rfds;
				FILE *f;

				((struct sockaddr_in *) &proxy)->sin_addr = a->proxy.sin_addr;
				((struct sockaddr_in *) &proxy)->sin_port = a->proxy.sin_port;
#if DEBUG
				fprintf(stderr, "magictun: connecting to proxy %d.%d.%d.%d:%d...\n",
						IPQUAD(a->proxy.sin_addr), ntohs(a->proxy.sin_port)
				);
#endif
				result = old_connect(sockfd, &proxy, sizeof(proxy));
				if (result < 0) return result;

				f = fdopen(dup(sockfd), "r+");
#if DEBUG
				fprintf(stderr, "magictun: sending 'CONNECT %d.%d.%d.%d:%d HTTP/1.0\r\n\r\n'\n",
						IPQUAD(a->orig.sin_addr), ntohs(a->orig.sin_port)
				);
#endif
				fprintf(f, "CONNECT %d.%d.%d.%d:%d HTTP/1.0\r\n\r\n",
						IPQUAD(a->orig.sin_addr), ntohs(a->orig.sin_port)
				);
				fflush(f);

				FD_ZERO(&rfds);
				FD_SET(sockfd, &rfds);
				tv.tv_sec = PROXY_TIMEOUT;
				tv.tv_usec = 0;

				if (select(sockfd + 1, &rfds, NULL, NULL, &tv)) {
					char buf[BUFSIZE];
					if (fgets(buf, sizeof(buf), f) == NULL) {
						fclose(f);
						return ENETUNREACH;
					}
#ifdef DEBUG
					fprintf(stderr, "magictun: got response: '%s'\n", buf);
#endif
					if (!strncmp(buf, "HTTP/1.0 200", 12)) {
#ifdef DEBUG
						fprintf(stderr, "magictun: ok\n");
#endif
						fgets(buf, sizeof(buf), f);
						fclose(f);
						return result;
					}
					else {
						fclose(f);
						return ENETUNREACH;
					}
				}
				else {
#ifdef DEBUG
					fprintf(stderr, "magictun: proxy timed out\n");
#endif
					fclose(f);
					return ETIMEDOUT;
				}
			}
		}
	}

	return old_connect(sockfd, serv_addr, addrlen);
}
